/********************************************************************
*               PROTOLOCOS DE ENLACE: GoBackN ARQ                   *
* ----------------------------------------------------------------- *
*      Membros: Anderson de Rezende Rocha - DCC/UFLA                *
*	            undersun@comp.ufla.br		                        *
*                                                                   *
*               Jlio Csar Alves - DCC/UFLA                        *
*	            jcalves@comp.ufla.br                                *
*                                                                   *    
*               Rodrigo Nazar da Silva Leite - DCC/UFLA            *
*	            canniggia@comp.ufla.br                              *
*                                                                   *
*               Fabiano Fernandes - DCC/UFLA                        *
*	            fernandes@comp.ufla.br                              *
*                                                                   *
* ----------------------------------------------------------------- *
*    REDES DE COMPUTADORES: Prof. Ricardo Martins de Abreu Silva    *
********************************************************************/

/* Incluso da biblioteca cnet responsvel pelo interpretao do 
   protocolo implementado 
*/ 
#include <cnet.h>
/* Biblioteca padro do C para manuseio da linguagem */ 
#include <stdlib.h>
/* Tratamento de strings */ 
#include <string.h>

/* Struct responsvel por representar a mensagem a ser transmitida
   e/ou recebida 
*/ 
typedef struct {
    char        dado[MAX_MESSAGE_SIZE];
} MENSAGEM;

/* Definindo os tipos que iro transitar pelo meio fsico
   Sero DADOS, RRs, ACKs e/ou SREJs 
   
   RR  do emissor e ACK  o RR do receptor 
*/ 
typedef enum    { DADOS, RR, ACK, SREJ }   TIPO_DE_QUADRO;

/* Struct responsvel por representar a quadro a ser transmitido
   e/ou recebido 
*/ 
typedef struct {
    TIPO_DE_QUADRO tipo;
    int tamanho;
    int checksum;
    int id;
    MENSAGEM mensagem;
} QUADRO;

/* Constante representando o tamanho do cabealho do quadro */ 
#define TAMANHO_CABECALHO_QUADRO  (sizeof(TIPO_DE_QUADRO) + 3*sizeof(int))
/* Constante representando o tamanho do quadro */ 
#define TAMANHO_QUADRO(q)      (TAMANHO_CABECALHO_QUADRO + q.tamanho)

/* Tamanho da janela desta abordagem.  passado em tempo de execuo */ 
static int tamanhoDaJanela;
/* Vetor representando os tamanhos das mensagens sendo 
   enviadas e/ou recebidas. 
   Do ponto de vista do 'receptor' representa quais as mensagens armazenada e que 
   no precisam ser reenviadas. 
   Do ponto de vista do 'emissor'  uma cpia para se acessar na necessidade de 
   retransmisso de algum quadro especfico
*/ 
static MENSAGEM *buffer;
/* Tamanho relativo a cada mensagem armazenada pelo buffer acima */
static int *tamanhoMensagensEnviadas;
/* 1 representa mensagem vlida no nosso buffer e 0 representa lixo 
   passvel de ser substitudo a qualquer momento */
static short int *statusDasMensagens; 
/* Representa a ltima mensagem enviada. Varivel de controle apenas */ 
static MENSAGEM ultimaMensagem;
/* Representa o ltimo tamanho ligado a uma mensagem enviada */
static int ultimoTamanho = 0;
/* Timer do cnet responsvel por cronometrar o tempo de envio/recebimento de 
   uma mensagem. til para verificar 'timeouts'
*/ 
static CnetTimer ultimoTimer = NULLTIMER;

/* Representa o identificador do prximo quadro a ser transmitido */ 
static int proximoQuadroId = 0;
/* Representa o identificador do quadro a ser recebido */ 
static int quadroIdEsperado = 0;
/* Representa o identificador do ltimo ACK enviado ao 'transmissor' */ 
static int ultimoACKEnviado;
/* Representa o identificador do ltimo ACK recebido pelo 'transmissor' */ 
static int ultimoACKRecebido;
/* Varivel dinmica que representa a 'rolagem' das janelas, ou seja, guarda
   o tamanho da janela assim que as situaes vo acontecendo
*/ 
static int tamanhoAtualDaJanela = 0;

/* Funo responsvel pelo controle ao se mudar o tamanho da janela */
static int incrementa(int n)
{
    n++;
    if(n > 2*tamanhoDaJanela)
        n = 0;
    return n;
}

/* Distncia representa quantos quadros existem entre o que n1 e n2 */
static int distancia(int n1, int n2)
{
    if (n2 >= n1)
        return (n2-n1);
    else
        return (2*tamanhoDaJanela-n1+n2+1);
}

/* Funo para DEBUG, ou melhor, acompanhamento do que est acontecendo */ 
static void debugAutomatico()
{
   int i = 0;

   printf(
    "\n\tproximoQuadroId\t= %d\n\tquadroIdEsperado\t= %d\n\ttamanhoAtualDaJanela\t= %d
    \n\tultimoACKEnviado\t= %d\n\tultimoACKRecebido\t= %d\n",
        proximoQuadroId, quadroIdEsperado, tamanhoAtualDaJanela,
        ultimoACKEnviado, ultimoACKRecebido);

    while(i < 2*tamanhoDaJanela+1)
    {
        printf("[%d] %d; ", i, statusDasMensagens[i]);
        i++;
    }
    printf("\n");
}

/* Funo utilizada tanto pelo 'transmissor' quanto pelo 'receptor' para transmitir
   um determinado quadro
*/
static void transmitirQuadro(MENSAGEM *mensagem, TIPO_DE_QUADRO tipo, int tamanhoDaMensagem, int idQuadro)
{
    /* Alocao do quadro a ser transmitido */
    QUADRO quadro;

    /* Atualizao de seus campos */
    quadro.tipo = tipo;
    quadro.id = idQuadro;
    quadro.checksum  = 0;
    quadro.tamanho = tamanhoDaMensagem;

    /* + Se o tipo for ACK e/ou SREJ  o 'receptor' que est enviando
       + Se o tipo for DADOS e/ou RR  o 'emissor' que est enviando
    */ 
    if(tipo == ACK)
    {
        printf("ACK transmitido, id = %d\n",idQuadro);
        tamanhoAtualDaJanela+= distancia(ultimoACKEnviado, quadro.id);
        ultimoACKEnviado = idQuadro;
    }
    else if (tipo == SREJ)
    {
        printf("SREJ transmitido, id = %d\n",idQuadro);
    }
    else if(tipo == DADOS)
    {
        /* Declarao de um 'timeout' a ser ligado  mensagem sendo 
           transmitida
        */
        CnetInt64 timeout;
        float f1;

        /* Copiando os dados para o quadro a ser transmitido */ 
        memcpy(&quadro.mensagem, (char *)mensagem, tamanhoDaMensagem);
        printf("DADOS transmitidos, id = %d\n",idQuadro);

        /* Contabilizando o tempo de propagao da mensagem 
           e a gerao da mesma
        */ 
        int64_L2F(f1, linkinfo[1].propagationdelay);
        int64_F2L(timeout, 1.0*(tamanhoDaJanela+1)*(f1 + 1000000*(TAMANHO_QUADRO(quadro) *
                    8.0)/linkinfo[1].bandwidth));

        /* Dando incio ao timer deste quadro */ 
        ultimoTimer = CNET_start_timer(EV_TIMER1, timeout, 0);
    }
    else if (tipo == RR)
    {
        /* Declarando o timeout */
        CnetInt64 timeout;
        float f1;

        printf("RR transmitido\n");
        /* Setando o tempo de propagao */
        int64_L2F(f1, linkinfo[1].propagationdelay);
        int64_F2L(timeout, 1.0*(tamanhoDaJanela+1)*(f1 + 1000000*(TAMANHO_QUADRO(quadro) *
                    8.0)/linkinfo[1].bandwidth));

        /* Dando incio ao timer deste quadro */ 
        ultimoTimer = CNET_start_timer(EV_TIMER1, timeout, 0);
    }

    /* Permitindo ao usurio acompanhar tudo que est acontecendo */ 
    debugAutomatico();
    /* Atualizando o tamannho da mensagem */ 
    tamanhoDaMensagem = TAMANHO_QUADRO(quadro);
    /* Utilizando o padro 'ccitt' para teste de checksum */
    quadro.checksum  = checksum_ccitt((unsigned char *)&quadro, tamanhoDaMensagem);
    /* Escrevendo a mensagem no meio fsico */ 
    CHECK(CNET_write_physical(1, (char *)&quadro, &tamanhoDaMensagem));
}


/* Se o 'emissor' pega um mensagem gerada pela camada de aplicao e faz o seu
   tratamento para posterior envio atravs do meio fsico 
*/
static void lerDaCamadaDeAplicacao(CnetEvent ev, CnetTimer timer, CnetData dados)
{
    int id;
    /* Endereo do destinatrio */
    CnetAddr enderecoDestino;
    if (tamanhoAtualDaJanela > 0)
    {
            /* Atualizando o tamanho da mensagem chegando */
            ultimoTamanho = sizeof(MENSAGEM);
            /* Lendo da camada de aplicao */
            CHECK(CNET_read_application(&enderecoDestino, (char *)&ultimaMensagem, &ultimoTamanho));
            printf("Despachado pela camada de aplicacao, id = %d\n", proximoQuadroId);
            /* Controlando no buffer a mensagem sendo lida da aplicao */
            id = proximoQuadroId;
            buffer[proximoQuadroId] = ultimaMensagem;
            tamanhoMensagensEnviadas[proximoQuadroId] = ultimoTamanho;
            tamanhoAtualDaJanela--;
            proximoQuadroId = incrementa(proximoQuadroId);
            /* Chamando funo prpria para o envio de mensagens */
            transmitirQuadro(&ultimaMensagem, DADOS, ultimoTamanho, id);
    }
    else
        CNET_disable_application(ALLNODES);
}

/* Utilizado tanto pelo 'receptor' para resgatar DADOS e RRs quanto pelo
   'emissor' para resgatar ACKs e REJs
*/
static void lerDaCamadaFisica(CnetEvent ev, CnetTimer timer, CnetData dados)
{
    /* Instanciando um quadro */ 
    QUADRO quadro;
    /* Tamanho representa o tamanho da mensagem
       Checksum representa o que chegou correto da mensagem segundo 
       o padro 'ccitt'
    */ 
    int link, tamanho, checksum;

    tamanho = sizeof(QUADRO);
    /* Lendo do meio fsico */ 
    CHECK(CNET_read_physical(&link,(char *)&quadro,&tamanho));

    /* Verificao do 'checksum' */ 
    checksum = quadro.checksum;
    quadro.checksum  = 0;
    /* Ignorando o quadro caso o 'checksum' seja invlido */ 
    if(checksum_ccitt((unsigned char *)&quadro, tamanho) != checksum)
    {
        printf("Checksum invalido - quadro ignorado\n");
        return;
    }

    /* Se o tipo de dados sendo lido  ACK quem o est lendo  o 'emissor' */
    if(quadro.tipo == ACK)
    {
        printf("ACK recebido, id = %d\n",quadro.id);
        if (quadro.id != ultimoACKRecebido)
            CNET_stop_timer(ultimoTimer);

        /* Calculando a distancia entre os quados para se atualizar o tamanho
           da janela
        */
        tamanhoAtualDaJanela += distancia(ultimoACKRecebido, quadro.id);
        ultimoACKRecebido = quadro.id;

        /* Retransmitindo uma mensagem em especfico que tenha sido seletivamente
           rejeita 'SREJ' pelo 'receptor' 
        */
        if (tamanhoAtualDaJanela == 0 || quadro.id != proximoQuadroId) /* verificar isto*/
        {
            printf("Despachado para retransmissao, id = %d\n", quadro.id);
            transmitirQuadro(&buffer[quadro.id], DADOS,
                             tamanhoMensagensEnviadas[quadro.id], quadro.id);
        }
        /* Habilitando a camada de aplicao a gerar novas mensagens */
        if (tamanhoAtualDaJanela != 0)
            CNET_enable_application(ALLNODES);
    }
    else if (quadro.tipo == SREJ)
    {
        /* Recebendo uma indicao de quadro rejeitado */
        if (quadro.id != proximoQuadroId)
        {
            printf("Indicacao de quadro rejeitado recebida, id = %d\n",quadro.id);
            printf("Despachado para retransmissao, id = %d\n", quadro.id);
            printf("tamanho retransmitido %d", tamanhoMensagensEnviadas[quadro.id]);
            /* Enviando quadro novamente */
            transmitirQuadro(&buffer[quadro.id], DADOS,
                              tamanhoMensagensEnviadas[quadro.id], quadro.id);
        }
        else
            printf("Quadro %d solicitado", quadro.id);
    }
    else if(quadro.tipo == DADOS)
    {
        /* Verifica se o 'recpetor' tem a capacidade de receber mensagens */
        if(tamanhoAtualDaJanela > 0 && distancia(quadroIdEsperado,quadro.id) < tamanhoDaJanela)
        {
            printf("\nDADOS recebidos, id = %d, ",quadro.id);
            if(quadro.id == quadroIdEsperado)
            {
                printf(" despachado para a camada de aplicacao\n");
                tamanho = quadro.tamanho;
                printf("tamanho passado %d", tamanho);
                /* Escrevendo na interface da camada de aplicao */
                CHECK(CNET_write_application((char *)&quadro.mensagem, &tamanho));

                /* Classificando a mensagem que chegou */
                statusDasMensagens[quadro.id] = 0;

                quadroIdEsperado = incrementa(quadroIdEsperado);
                /* Mensagem recebida casa com mensagem esperada, enviando para camada
                   de aplicao
                */
                while(statusDasMensagens[quadroIdEsperado] == 1)
                {
                    printf(" despachado para a camada de aplicacao, id = %d\n", quadroIdEsperado);
                    tamanho = tamanhoMensagensEnviadas[quadroIdEsperado];
                    /* Escrevendo na camada de aplicao um conjunto de mensagens j ordenados
                       no buffer
                    */
                    CHECK(CNET_write_application((char *)&buffer[quadroIdEsperado], &tamanho));
                    /* Setando o status das mensagens para 0 */
                    statusDasMensagens[quadroIdEsperado] = 0;
                    quadroIdEsperado = incrementa(quadroIdEsperado);
                }
                /* Verificando um caso de exceo */
                if(tamanhoAtualDaJanela == 0)
                    printf("DEBUG, nao poderia receber\n");
                else
                    tamanhoAtualDaJanela--;

                /* Gerando um nmero aleatrio para verificar se o 'receptor' deve
                   ou no transmitir um ACK neste momento
                */
                if ( (int)(rand()%3) == 0 )
                    transmitirQuadro((MENSAGEM *)NULL, ACK, 0, quadroIdEsperado);
            }
            else
            {
                if(tamanhoAtualDaJanela == 0)
                    printf("DEBUG, nao poderia receber\n");
                else
                    tamanhoAtualDaJanela--;

                /* Mensagem recebida correta mas fora de ordem. Armazena-se no buffer */
                tamanho = quadro.tamanho;
                buffer[quadro.id] = quadro.mensagem;
                tamanhoMensagensEnviadas[quadro.id] = tamanho;
                statusDasMensagens[quadro.id] = 1;
                printf("Quadro %d armazenado no buffer\n", quadro.id);
                transmitirQuadro((MENSAGEM *)NULL, SREJ, 0, quadroIdEsperado);
            }
        }
        else if(tamanhoAtualDaJanela == 0)
        {
            /* Rejeitando-se seletivamente uma determinada mensagem com erro detectado */
            printf("Mensagem recebida ignorada, buffer cheio. Id = %d", quadro.id);
            transmitirQuadro((MENSAGEM *)NULL, SREJ, 0, quadroIdEsperado);
        }
        else
        {
            printf("Mensagem recebida ignorada, fora do intervalo da janela. Id = %d\n", quadro.id);
            transmitirQuadro((MENSAGEM *)NULL, SREJ, 0, quadroIdEsperado);
        }
    }
    else /* quadro.tipo == RR */
    {
        /* 'Receptor' recebeu uma 'pergunta'. Respondendo, ento, com um 
            ACK indicando qual o prximo quadro esperado para recebimento
        */
        printf("RR recebido\n");
        transmitirQuadro((MENSAGEM *)NULL, ACK, 0, quadroIdEsperado);
    }
}

/* Desejando na 'tela' do simulador os quadros transitando pelo meio fsico
   bem como as aes que este meio impem aos mesmos
*/
static void draw_quadro(CnetEvent ev, CnetTimer timer, CnetData data)
{
    CnetDrawFrame *df	= (CnetDrawFrame *)data;
    QUADRO	  *f	= (QUADRO *)df->frame;

    if(f->tipo == ACK || f->tipo == SREJ) {
	df->colour[0]	= (f->id == 0) ? CN_RED : CN_PURPLE;
	df->pixels[0]	= 10;
    }
    else if(f->tipo == DADOS || f->tipo == RR) {
	df->colour[0]	= (f->id == 0) ? CN_RED : CN_PURPLE;
	df->pixels[0]	= 10;
	df->colour[1]	= CN_GREEN;
	df->pixels[1]	= 30;
    }
}

/* Verifica se um determinado quadro atingiu 'timeout' ,ou melhor, 
   se um determinado quadro transmitido estourou seu tempo inerente 
   de transmisso
*/
static void timeouts(CnetEvent ev, CnetTimer timer, CnetData data)
{
    if(timer == ultimoTimer)
    {
        printf("Timeout atingido\n");
        transmitirQuadro((MENSAGEM *)NULL,RR,0,ultimoACKRecebido);
    }
}

/* Mostra o estado interno deste protocolo. til para que o usurio acompanhe tudo 
   que est acontecendo
*/
static void showstate(CnetEvent ev, CnetTimer timer, CnetData data)
{
    printf(
    "\n\tproximoQuadroId\t= %d\n\tquadroIdEsperado\t= %d\n\ttamanhoAtualDaJanela\t= %d
    \n\tultimoACKEnviado\t= %d\n\tultimoACKRecebido\t= %d\n",
        proximoQuadroId, quadroIdEsperado, tamanhoAtualDaJanela,
        ultimoACKEnviado, ultimoACKRecebido);
}

/* Funo principal que  chamada pelo simulador para dar-se incio  simulao
   do protocolo escrito 
*/
void reboot_node(CnetEvent ev, CnetTimer timer, CnetData data)
{
    char **argv = (char **)data;

    /* Se nenhum parmetro foi passado no se pode proceder a simulao */ 
    if(argv[0])
    {
        tamanhoDaJanela = atoi(argv[0]);
        printf("Tamanho da janela passado: %d\n", tamanhoDaJanela);
        if(tamanhoDaJanela != 0)
        {
            if(nodeinfo.nodenumber > 1)
            {
                /* Se o arquivo de topologia no descreve uma rede de 2 ns no se
                   pode prosseguir
                */
                fprintf(stderr,"Isto no  uma rede de 2 ns!\n");
                exit(1);
            }
            /* Inicializao dos parmetros dinmicos a serem utilizados pelo protocolo */
            tamanhoAtualDaJanela = tamanhoDaJanela;
            buffer = calloc(2*tamanhoDaJanela+1, sizeof(MENSAGEM));
            tamanhoMensagensEnviadas = calloc(2*tamanhoDaJanela+1, sizeof(int));
            statusDasMensagens = calloc(2*tamanhoDaJanela+1, sizeof(short int));
            ultimoACKEnviado = 0;
            ultimoACKRecebido = 0;

            /* Declarao dos eventos do CNET ligados  simulao */
            CHECK(CNET_set_handler( EV_APPLICATIONREADY, lerDaCamadaDeAplicacao, 0));
            CHECK(CNET_set_handler( EV_PHYSICALREADY,lerDaCamadaFisica, 0));
            CHECK(CNET_set_handler( EV_DRAWFRAME, draw_quadro, 0));
            CHECK(CNET_set_handler( EV_TIMER1, timeouts, 0));
            CHECK(CNET_set_handler( EV_DEBUG1, showstate, 0));

            CHECK(CNET_set_debug_string( EV_DEBUG1, "Estado"));

            /* Dando incio  gerao de mensagens */
            if(nodeinfo.nodenumber == 1)
                CNET_enable_application(ALLNODES);
        }
        else
            printf("Tamanho da janela passado invlido.\nA simulao no ser executada.");
    }
    else
        printf("Voc deve passar o tamanho da janela na linha de comando.\n$ cnet <arquivo de topologia>
                <tamanho da janela>");
}
